from typing import Final

# These are not actually bools. See #4669
NO: Final[bool]
YES: Final[bool]
TRUE: Final[bool]
FALSE: Final[bool]
ON: Final[bool]
OFF: Final[bool]
N: Final = "n"
S: Final = "s"
W: Final = "w"
E: Final = "e"
NW: Final = "nw"
SW: Final = "sw"
NE: Final = "ne"
SE: Final = "se"
NS: Final = "ns"
EW: Final = "ew"
NSEW: Final = "nsew"
CENTER: Final = "center"
NONE: Final = "none"
X: Final = "x"
Y: Final = "y"
BOTH: Final = "both"
LEFT: Final = "left"
TOP: Final = "top"
RIGHT: Final = "right"
BOTTOM: Final = "bottom"
RAISED: Final = "raised"
SUNKEN: Final = "sunken"
FLAT: Final = "flat"
RIDGE: Final = "ridge"
GROOVE: Final = "groove"
SOLID: Final = "solid"
HORIZONTAL: Final = "horizontal"
VERTICAL: Final = "vertical"
NUMERIC: Final = "numeric"
CHAR: Final = "char"
WORD: Final = "word"
BASELINE: Final = "baseline"
INSIDE: Final = "inside"
OUTSIDE: Final = "outside"
SEL: Final = "sel"
SEL_FIRST: Final = "sel.first"
SEL_LAST: Final = "sel.last"
END: Final = "end"
INSERT: Final = "insert"
CURRENT: Final = "current"
ANCHOR: Final = "anchor"
ALL: Final = "all"
NORMAL: Final = "normal"
DISABLED: Final = "disabled"
ACTIVE: Final = "active"
HIDDEN: Final = "hidden"
CASCADE: Final = "cascade"
CHECKBUTTON: Final = "checkbutton"
COMMAND: Final = "command"
RADIOBUTTON: Final = "radiobutton"
SEPARATOR: Final = "separator"
SINGLE: Final = "single"
BROWSE: Final = "browse"
MULTIPLE: Final = "multiple"
EXTENDED: Final = "extended"
DOTBOX: Final = "dotbox"
UNDERLINE: Final = "underline"
PIESLICE: Final = "pieslice"
CHORD: Final = "chord"
ARC: Final = "arc"
FIRST: Final = "first"
LAST: Final = "last"
BUTT: Final = "butt"
PROJECTING: Final = "projecting"
ROUND: Final = "round"
BEVEL: Final = "bevel"
MITER: Final = "miter"
MOVETO: Final = "moveto"
SCROLL: Final = "scroll"
UNITS: Final = "units"
PAGES: Final = "pages"
