class F[T](y := list): ...
class I[T]((yield 1)): ...
class J[T]((yield from 1)): ...
class K[T: (yield 1)]: ...      # yield in TypeVar
class L[T: (x := 1)]: ...       # named expr in TypeVar
class M[T]((await 1)): ...
class N[T: (await 1)]: ...
